'''
Copyright (C) 2019 JOSECONSCO
Created by JOSECONSCO

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

bl_info = {
    "name": "Volume Preserving Smoothing",
    "description": "Volume preserving Smoothing",
    "author": "Bartosz Styperek",
    "version": (2, 8, 0),
    "blender": (3, 6, 0),
    "location": "In Mesh Edit Mode: W-key -> Volume Smooth. Or use Volume Smooth tool from left tool sidebar ",
    "warning": "",
    "tracker_url": "https://discord.gg/cxZDbqH",
    "category": "Mesh",
}

if "bpy" in locals():
    import importlib
    importlib.reload(vps_prefs)
    importlib.reload(volume_preserve_smoothing)
    importlib.reload(update_addon)
    importlib.reload(smooth_tool)

else:
    from . import vps_prefs
    from . import volume_preserve_smoothing
    from . import update_addon
    from . import smooth_tool

import bpy

from . import auto_load
auto_load.init()



def smooth_menu_last(self, context):
    layout = self.layout
    layout.separator()
    layout.menu('VPS_MT_SmoothMenu')

def register():
    auto_load.register()

    bpy.types.Scene.vps_gizmo_show = bpy.props.BoolProperty(default = True)
    bpy.types.Object.vps = bpy.props.PointerProperty(type=vps_prefs.VPS_PropertyGroup)
    print("Registered Volume Preserving Smoothing")
    addon_prefs = update_addon.get_addon_preferences()
    if addon_prefs:
        addon_prefs.update_text = ''
    bpy.types.VIEW3D_MT_edit_mesh_context_menu.append(smooth_menu_last)
    bpy.types.VIEW3D_MT_edit_mesh_vertices.append(smooth_menu_last)
    bpy.utils.register_tool(smooth_tool.ToolVpsSmooting, after={"builtin.smooth"}, separator=True, group=True)



def unregister():
    bpy.utils.unregister_tool(smooth_tool.ToolVpsSmooting)

    bpy.types.VIEW3D_MT_edit_mesh_context_menu.remove(smooth_menu_last)
    bpy.types.VIEW3D_MT_edit_mesh_vertices.remove(smooth_menu_last)
    auto_load.unregister()
    print("Unregistered Volume Preserving Smoothing")
    del bpy.types.Scene.vps_gizmo_show
    del bpy.types.Object.vps


if __name__ == "__main__":
    register()
